import torch
import torch.nn as nn
from smolhub.helper.dataset.load_config import Config
from dataclasses import dataclass

config = Config().get_config()

@dataclass
class ModelArgs:
    device: str = torch.device("cuda" if torch.cuda.is_available() else "cpu") 
    rank = config["LoRA"]["rank"]
    alpha = config["LoRA"]["alpha"]



class LoRALayer(nn.Module):
    def __init__(self, model) -> None:
        super().__init__()


        self.rank = ModelArgs.rank
        self.model_weight_dims = model.config.n_embd
        self.query_A = nn.Parameter(torch.ones((self.model_weight_dims, self.rank), requires_grad=True))
        self.query_B = nn.Parameter(torch.zeros((self.rank, self.model_weight_dims), requires_grad=True))
        self.key_A = nn.Parameter(torch.ones((self.model_weight_dims, self.rank), requires_grad=True))
        self.key_B = nn.Parameter(torch.zeros((self.rank, self.model_weight_dims), requires_grad=True))
        self.value_A = nn.Parameter(torch.ones((self.model_weight_dims, self.rank), requires_grad=True))
        self.value_B = nn.Parameter(torch.zeros((self.rank, self.model_weight_dims), requires_grad=True))
        self.output_A = nn.Parameter(torch.ones((self.model_weight_dims, self.rank), requires_grad=True))
        self.output_B = nn.Parameter(torch.zeros((self.rank, self.model_weight_dims), requires_grad=True))
        # self.linear_q = nn.Linear(in_features=model.config.n_ctx, out_features=self.model_weight_dims, bias=False)
        # self.linear_k = nn.Linear(in_features=model.config.n_ctx, out_features=self.model_weight_dims, bias=False)
        # self.linear_v = nn.Linear(in_features=model.config.n_ctx, out_features=self.model_weight_dims, bias=False)
        # self.linear_o = nn.Linear(in_features=model.config.n_ctx, out_features=self.model_weight_dims, bias=False)
        torch.nn.init.normal_(self.query_A, mean=0.0, std=1)
        torch.nn.init.normal_(self.key_A, mean=0.0, std=1)
        torch.nn.init.normal_(self.output_A, mean=0.0, std=1)
        torch.nn.init.normal_(self.value_A, mean=0.0, std=1)


    def forward(self, w_o, q_o, k_o, v_o):
        # print((self.output_B).shape)
        final_weight_WO = w_o + self.output_B.T @ self.output_A.T
        final_weight_QO = q_o + self.query_B.T @ self.query_A.T
        final_weight_KO = k_o + self.key_B.T @ self.key_A.T
        final_weight_VO = v_o + self.value_B.T @ self.value_A.T
        # out_q = self.linear_q(final_weight_QO)
        # out_k = self.linear_k(final_weight_KO)
        # out_v = self.linear_v(final_weight_VO)

        # out_o
        #
        #
        # = self.linear_o(final_weight_WO)

        return final_weight_WO, final_weight_QO , final_weight_KO, final_weight_VO


class LoRAWrapper(nn.Module):
    def __init__(self, model) -> None:
        super().__init__()
        self.lora_layer = LoRALayer(model)
        # self.linear = nn.Linear(in_features=model.config.vocab_size, out_features=2)
        self.config = model.config
        self.model = model
    def forward(self, x):
        qkv_layers = [self.model.transformer.h[i].attn.c_attn for i in range(self.config.n_layer)]
        o_layers = [self.model.transformer.h[i].attn.c_proj for i in range(self.config.n_layer)]

        for i in range(len(qkv_layers)):
            hidden_size = qkv_layers[i].weight.size(-1) // 3
            Q, K, V = torch.split(qkv_layers[i].weight, hidden_size, dim=-1)
            O = o_layers[i].weight
            out_o, out_q, out_k, out_v = self.lora_layer(O,Q,K,V)
            combined_qkv = torch.concat([out_q, out_k, out_v], dim=-1)
            # print(combined_qkv.shape)
            # Update the model's attention weights
            # # with torch.no_grad():
            # qkv_layers[i].weight.copy_(combined_qkv)
            # o_layers[i].weight.copy_(out_o)
            # Assign the updated weights back to the model
            qkv_layers[i].weight.data.copy_(combined_qkv)
            o_layers[i].weight.data.copy_(out_o)
        return self.model(x)




class LoRAModel(nn.Module):
    def __init__(self, model) -> None:
        super().__init__()
        
        # Freeze base model parameters
        for param in model.parameters():
            param.requires_grad = False
            
        self.lora_wrapper = LoRAWrapper(model)
        self.config = model.config

    def forward(self, x):

        out = self.lora_wrapper(x)

        return out




import yaml


class Config:
    def __init__(self, config_path='/mnt/c/Users/yuvra/OneDrive/Desktop/Work/pytorch/SmolHub/src/smolhub/config/config.yaml'):
        self.config_path = config_path
        self.config = self.load_config()

    def load_config(self):
        with open(self.config_path, "r") as f:
            config = yaml.safe_load(f)
        return config

    def get_config(self):
        return self.config




import torch
from torchinfo import summary

from smolhub.scripts.lora import ModelArgs

def print_model(model, dataloader):
    # lora_model = LoRAModel()
    model.to(ModelArgs.device)
    # input_data = next(iter(dataloader))
    input_ids = torch.randint(
    0, model.config.vocab_size,
    (1, model.config.n_ctx)
).to(ModelArgs.device)
    #Printing a summary of the architecture
    
    summary(model=model,
            input_data=input_ids,
            col_names=["input_size", "output_size", "num_params", "trainable"],
            col_width=20,
            row_settings=["var_names"])




import math

class CustomLRScheduler:
    def __init__(self, optimizer, warmup_iters, lr_decay_iters, min_lr, max_lr, _type):
        self.optimizer = optimizer
        self.warmup_iters = warmup_iters
        self.lr_decay_iters = lr_decay_iters
        self.min_lr = min_lr
        self.max_lr = max_lr
        self.it = 0
        self._last_lr = [max_lr]  # Initialize with max_lr (matching PyTorch convention)
        self.type = _type

    def step(self):
        
        self._last_lr = [self._get_lr()]  # Store as list
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = self._last_lr[0]
        self.it += 1

    def get_last_lr(self):
        return self._last_lr  # Returns list to match PyTorch convention
    
    def _get_lr(self):

        if(self.type == 'cycle'):
            cycle = math.floor(1 + self.it / (2 * self.warmup_iters))
            x = abs(self.it / self.warmup_iters - 2 * cycle + 1)
            return self.min_lr + (self.max_lr - self.min_lr) * max(0, (1 - x))

        elif(self.type == 'cosine'):
            # 1) linear warmup for warmup_iters steps
            if self.it < self.warmup_iters:
                return self.max_lr * (self.it + 1) / (self.warmup_iters + 1)
            # 2) if it > lr_decay_iters, return min learning rate
            if self.it > self.lr_decay_iters:
                return self.min_lr
            # 3) in between, use cosine decay down to min learning rate
            decay_ratio = (self.it - self.warmup_iters) / (self.lr_decay_iters - self.warmup_iters)
            assert 0 <= decay_ratio <= 1
            coeff = 0.5 * (1.0 + math.cos(math.pi * decay_ratio)) 
            return self.min_lr + coeff * (self.max_lr - self.min_lr)
    
    def state_dict(self):
        return {
            'warmup_iters': self.warmup_iters,
            'lr_decay_iters': self.lr_decay_iters,
            'min_lr': self.min_lr,
            'max_lr': self.max_lr,
            'it': self.it
        }
    
    def load_state_dict(self, state_dict):
        self.warmup_iters = state_dict['warmup_iters']
        self.lr_decay_iters = state_dict['lr_decay_iters']
        self.min_lr = state_dict['min_lr']
        self.max_lr = state_dict['max_lr']
        self.it = state_dict['it']



import torch
import smolhub
# from smolhub.helper.dataset.load_config import Config
from smolhub.scripts.finetune import SFTTrainer
from transformers import AutoTokenizer, AutoModelForCausalLM
from smolhub.helper.scheduler import CustomLRScheduler
from smolhub.scripts.lora import LoRAModel
from smolhub.helper.dataset.dataset_main import PreprocessDataset
from load_config import Config #Needs to be created

model_id = "openai-community/gpt2"

config = Config().get_config()
dataset_path = config["Dataset"]["dataset_path"]

tokenizer = AutoTokenizer.from_pretrained(model_id, token=config['huggingface']['hf_token'])
model = AutoModelForCausalLM.from_pretrained(model_id, device_map="auto", token=config['huggingface']['hf_token'])

if tokenizer.pad_token is None:
    # Set the pad token to the eos token if it doesn't exist
    tokenizer.add_special_tokens({'pad_token': '[PAD]'})
    # tokenizer.pad_token = tokenizer.eos_token
    

    print("Setting pad token as PAD token ")

model.resize_token_embeddings(len(tokenizer))

lora_model = LoRAModel(model)
optimizer = torch.optim.Adam(lora_model.parameters(), lr=2e-3)
scheduler = CustomLRScheduler(optimizer, warmup_iters=100, lr_decay_iters=2000, min_lr=2e-5, max_lr=2e-3, _type="cosine")

#Loading the dataset
preprocess_dataset = PreprocessDataset(dataset_path=dataset_path, tokenizer=tokenizer)
train_dataloader, val_dataloader, test_dataloader = preprocess_dataset.prepare_dataset()

#Initialize the Trainer
sft_trainer = SFTTrainer(lora_model, train_dataloader, val_dataloader, test_dataloader, optimizer, None, scheduler)

#Train
sft_trainer.train()





from datasets import load_dataset

from smolhub.helper.dataset.prepare_dataset import SFTDataset, convert_dataset_to_dataloader
from smolhub.helper.dataset.load_config import Config

config = Config().get_config()

class PreprocessDataset:
    def __init__(self, dataset_path=None, tokenizer=None):
        self.dataset_path = dataset_path
        self.use_hf_dataset = config["Dataset"]["use_hf_dataset"]
        # self.tokenizer_path = config["Model"]["tokenizer"]
        self.train = None
        self.val = None
        self.test = None
        self.tokenizer = tokenizer
       
        
    def check_for_split(self, dataset):
        # Initialize all splits to None
        self.train = self.val = self.test = None
        
        # Case 1: Dataset has all three splits
        if all(split in dataset for split in ['train', 'validation', 'test']):
            print("Dataset splits found: train, validation, test")
            self.train = dataset['train']
            self.val = dataset['validation']
            self.test = dataset['test']
            print(self.train)
            print(self.val)
            print(self.test)
            return
        
        # Case 2: Only train exists - create both val and test (10% each)
        if 'train' in dataset and 'validation' not in dataset and 'test' not in dataset:
            print("Creating validation and test splits from train")
            # First split train into temp_train (90%) and test (10%)
            temp_split = dataset['train'].train_test_split(test_size=0.1)
            # Then split temp_train into final train (80%) and val (10%)
            final_split = temp_split['train'].train_test_split(test_size=1/9)  # 0.1/0.9 ≈ 0.111
            self.train = final_split['train']
            self.val = final_split['test']
            self.test = temp_split['test']
            print(self.train)
            print(self.val)
            print(self.test)
            return
        
        # Case 3: Has train and test but no validation - split train to create validation
        if 'train' in dataset and 'test' in dataset and 'validation' not in dataset:
            print("Creating validation split from train")
            self.test = dataset['test']
            # Split train into new train (90%) and validation (10%)
            split = dataset['train'].train_test_split(test_size=0.1)
            self.train = split['train']
            self.val = split['test']
            print(self.train)
            print(self.val)
            print(self.test)
            return
        
        # Case 4: Has train and validation but no test - split train to create test
        if 'train' in dataset and 'validation' in dataset and 'test' not in dataset:
            print("Creating test split from train")
            self.val = dataset['validation']
            # Split train into new train (90%) and test (10%)
            split = dataset['train'].train_test_split(test_size=0.1)
            self.train = split['train']
            self.test = split['test']
            print(self.train)
            print(self.val)
            print(self.test)
            return
        
        # Case 5: Only test exists (unlikely but possible) - use test as validation and split to create new test
        if 'test' in dataset and 'train' not in dataset:
            print("Warning: Only test split found. Using half as validation and splitting to create new test")
            # Split test into val and new test (50% each)
            split = dataset['test'].train_test_split(test_size=0.5)
            self.val = split['train']
            self.test = split['test']
            print(self.train)
            print(self.val)
            print(self.test)
            # No train data available in this case
            return
        
        # Case 6: Only validation exists (unlikely but possible) - use as test and split to create new validation
        if 'validation' in dataset and 'train' not in dataset:
            print("Warning: Only validation split found. Using as test and splitting to create new validation")
            # Split validation into val and test (50% each)
            split = dataset['validation'].train_test_split(test_size=0.5)
            self.val = split['train']
            self.test = split['test']
            print(self.train)
            print(self.val)
            print(self.test)
            # No train data available in this case
            return
        
        # If we get here, the dataset has an unexpected combination of splits
        raise ValueError("Unexpected dataset split configuration. Could not create required splits.")
        
        # A full HF dataset object is required rest it will take care
    def prepare_dataset(self):
        
        if(self.use_hf_dataset):
            self.dataset = load_dataset(self.dataset_path, token=config["huggingface"]["hf_token"])
            # print(self.dataset)
            self.check_for_required_columns(self.dataset)
            self.labels = self.get_unique_labels(self.dataset)
            self.check_for_split(self.dataset)
            self.check_for_required_columns(self.train)
            self.check_for_required_columns(self.test)
            self.check_for_required_columns(self.val)
            
            
            # self.train = self.train.map(self.tokenize)
            # self.val = self.val.map(self.tokenize)
            # self.test = self.test.map(self.tokenize)
            
            # self.train = self.train.remove_columns(["text"])
            # self.val = self.val.remove_columns(["text"])
            # self.test = self.test.remove_columns(["text"])
            
            train_dataloader = convert_dataset_to_dataloader(self.train, self.tokenizer, self.labels)
            val_dataloader = convert_dataset_to_dataloader(self.val, self.tokenizer , self.labels)
            test_dataloader = convert_dataset_to_dataloader(self.test, self.tokenizer   , self.labels)
            
            return train_dataloader, val_dataloader, test_dataloader   
        # If not using HF dataset, we are using a custom dataset
        # Just a full csv or json is required with train key value pair
        else:
            self.sft_dataset = SFTDataset(self.dataset_path)
            self.check_for_required_columns(self.sft_dataset)
            self.labels = self.get_unique_labels(self.sft_dataset)
            self.check_for_split(self.sft_dataset)
            self.check_for_required_columns(self.train)
            self.check_for_required_columns(self.test)
            self.check_for_required_columns(self.val)
            
            # self.train = self.train.map(self.tokenize)
            # self.val = self.val.map(self.tokenize)
            # self.test = self.test.map(self.tokenize)
            
            train_dataloader = convert_dataset_to_dataloader(self.train, self.tokenizer, self.labels)
            val_dataloader = convert_dataset_to_dataloader(self.val, self.tokenizer, self.labels)
            test_dataloader = convert_dataset_to_dataloader(self.test, self.tokenizer, self.labels)
            
            return train_dataloader, val_dataloader, test_dataloader

    
    
    def check_for_required_columns(self, dataset):
        required_columns = ["text", "label"]
        for column in required_columns:
            if column not in dataset.column_names:
                raise ValueError(f"Required column {column} not found in dataset")
   
    def check_label_column(self, dataset):
        if dataset['label'].dtype != 'float32' or dataset['label'].dtype != 'int32':
            raise ValueError("Label column must be of type float or int")


    def get_unique_labels(self, dataset):
        labels = []
        for example in dataset['train']:
            # print(example)
            if example['label'] not in labels:
                labels.append(example['label'])
        return labels
        



import torch
from smolhub.helper.dataset.load_config import Config

from torch.cuda.amp import GradScaler
from smolhub.helper.count_parameters import count_parameters
from tqdm import tqdm
from smolhub.helper.visualize import Visualizer
# from tests.print_model_details import print_model




class SFTTrainer:
    def __init__(self, model, train_dataloader, val_dataloader, test_dataloader, optimizer, loss_fn, scheduler=None):

        self.model = model
        self.train_dataloader = iter(train_dataloader)
        self.val_dataloader = iter(val_dataloader)
        self.test_dataloader = iter(test_dataloader)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        
        self.config = Config().get_config()
        self.visualizer = Visualizer()  

        # self.device = 'cpu'
        self.model.to(self.device)
        self.optimizer = optimizer
        self.loss_fn = loss_fn
        self.epochs =  self.config["Model"]["epochs"]
        # self.model.train()
        self.eval_iters =  self.config["Model"]["eval_iters"]
        
        #None cus there is loss mask to multiply it with
        self.loss_fn = torch.nn.CrossEntropyLoss(reduction='none')
        
        #Getting model details and showing to the user
        # print_model(self.model, self.train_dataloader)
      
        print("Total trainable parameters:", count_parameters(self.model) ," which is: " , (count_parameters(self.model) / 163037184 )*100 , "%\ of" , 163037184 , "trainable params")
        print("\n")

        # Mixed Precision Training
        self.scaler = GradScaler(enabled=( self.config["MAP"]["use_float16"]))

        if( self.config["Optimizations"]["use_compile"]):
            self.model = torch.compile(self.model)
        #load scheduler
        self.scheduler = scheduler


            
    @torch.inference_mode()
    def evaluate(self):
        
        out = {}
        self.model.eval()

        for split in ['val']:
            losses = torch.zeros(eval_iters := len(self.val_dataloader), device=self.device)
            for k in range(len(self.val_dataloader)):
                with torch.autocast(device_type='cuda', dtype=torch.bfloat16 if  self.config["MAP"]["use_bfloat16"] or  self.config["MAP"]["use_float16"] else torch.float32):
                    batch= next(self.train_dataloader)
                    idx = batch['input_ids'].to(self.device)
                    targets = batch['labels'].to(self.device)
                    loss_mask = batch['loss_mask'].to(self.device)
                    logits = self.model(idx).logits
                    batch_size, block_size, embeddings_dims = logits.shape
                    logits = logits.view(batch_size*block_size, embeddings_dims)
                    targets = targets.view(batch_size * block_size)
                    loss = self.loss_fn(logits, targets)
                loss = loss * loss_mask.view(-1)
                loss = loss.mean()
                # print('Val: ', loss.item())
            out[split] = losses.mean()

        self.model.train()
        return out

    def train(self):
        
        # self.model.train()
        for epoch in range(self.epochs):
            for step in tqdm(range(len(self.train_dataloader))):

                self.total_steps = len(self.train_dataloader) * self.epochs
                if (step  % self.eval_iters == 0 and step != 0) or step == self.total_steps - 1:
                    losses = self.evaluate()
                    print(f"epoch {epoch}, step {step}: train loss {losses['train']:.4f}, val loss {losses['val']:.4f}")
                    self.visualizer.log({
                        "val_loss": losses['val'],
                        
                    })
                self.optimizer.zero_grad(set_to_none=True)
                with torch.autocast(device_type='cuda', dtype=torch.bfloat16 if  self.config["MAP"]["use_bfloat16"] or  self.config["MAP"]["use_float16"] else torch.float16):
                    batch= next(self.train_dataloader)
                    idx = batch['input_ids'].to(self.device)
                    targets = batch['labels'].to(self.device)
                    loss_mask = batch['loss_mask'].to(self.device)
                    
                    logits = self.model(idx).logits
                    # print(logits)
                    batch_size, block_size, embeddings_dims = logits.shape
                    logits = logits.view(batch_size*block_size, embeddings_dims)
                    targets = targets.view(batch_size * block_size)
                    loss = self.loss_fn(logits, targets)
                    # print(loss_mask)
                    # print(idx)
                    # print(targets)
                    loss = loss * loss_mask.view(-1)
                    loss = loss.mean()
                
                print('Train: ', loss.item())
                loss.requires_grad_(True)
                if  self.config["MAP"]["use_float16"]:
                    self.scaler.scale(loss).backward()
                    # Unscale gradients and optionally clip them
                    # self.scaler.unscale_(self.optimizer)
                  
                    self.scaler.step(self.optimizer)
                    self.scaler.update()
                else:
                    loss.backward()
                   
                    self.optimizer.step()
                    # self.optimizer.step()
                self.scheduler.step()

                self.visualizer.log({
                    "epoch": epoch,
                    "step": step,
                    "train_loss": loss.item(),
                    "lr": self.scheduler.get_last_lr()[0],
                    # 'grad_norm': total_norm_before.item(),
                })


from setuptools import setup, find_packages

setup(
    name="smolhub",
    version="0.5.1",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    package_data={
        "smolhub": ["config/*.yaml"],
    },
    install_requires=[
        "torch",
        "transformers",
        "datasets",
        "wandb",
        "tqdm",
        "pyyaml",
        "numpy",
    ],
    author="Yuvraj Singh",
    author_email="yuvraj.mist@gmail.com",
        long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    include_package_data=True,
)


import torch
from torch.utils.data import Dataset
import pandas as pd
from datasets import Dataset
from smolhub.helper.dataset.load_config import Config
import torch
from torch.nn.utils.rnn import pad_sequence
from transformers import BatchEncoding


config = Config().get_config()

def json_to_csv(json_dataset):
    # Convert JSON dataset to pd.DataFrame
    csv_dataset = pd.DataFrame(json_dataset)
    return csv_dataset

def _load_dataset(dataset_path):
    # Load dataset from the given path and convert it to a HF Dataset
    if(dataset_path.endswith(".csv")):
        dataset = pd.read_csv(dataset_path)
        hf_dataset = Dataset.from_pandas(dataset)
        return hf_dataset
    elif(dataset_path.endswith(".json")):
        dataset = pd.read_json(dataset_path)
        csv_dataset = json_to_csv(dataset)
        hf_dataset = Dataset.from_pandas(csv_dataset)
        return hf_dataset
    elif(dataset_path.endswith(".xlsx")):
        dataset = pd.read_excel(dataset_path)
        hf_dataset = Dataset.from_pandas(dataset)
        return hf_dataset
    else:
        raise ValueError("Unsupported dataset format")




    
    

def sft_collate_fn(batch_samples, tokenizer, config, labels):
    """
    Collate function for SFTDataset that:
    1. Handles dynamic padding
    2. Applies loss masking (only compute loss on label portion)
    3. Prepares inputs for causal language modeling
    """
    batch = {
        'input_ids': [],
        
        'labels': [],
        'loss_mask': []
    }
    
    for sample in batch_samples:
        # print(sample)
        text = sample['text']
        label = sample['label']
        
            
        # Format the prompt based on task type
        if config['Dataset']['type'] == "classification":
            
            cls_format = """
                Here is a text:
                Text: {}

                You are tasked to do classification.
                
                
                SOLUTION
                The correct answer is: {}
                
                """
            full_prompt = cls_format.format(text, label)
            prompt_before_label = cls_format.format(text, "")
            # print(full_prompt)
            # print(prompt_before_label)
        else:
            raise ValueError("Unsupported dataset type")

        # Tokenize full prompt and the part before label
        tokenized_text = tokenizer(
            full_prompt,
            truncation=True,
            # padding='max_length',
            # max_length=config["Dataset"]["max_length"],
            return_tensors='pt'
        )
        
        tokenized_prompt_before_label = tokenizer(
            prompt_before_label,
            truncation=True,
            # padding='max_length',
            # max_length=config["Dataset"]["max_length"],
            return_tensors='pt'
        )

        # Calculate where the label starts
        label_start_idx = len(tokenized_prompt_before_label['input_ids'])
        
        # Create loss mask (1 for label tokens, 0 otherwise)
        loss_mask = torch.zeros_like(tokenized_text['input_ids'])
        loss_mask[:, label_start_idx:] = 1
        
        # Prepare labels for causal LM (shifted right)
        labels = tokenized_text['input_ids'].clone()
        labels[:, :-1] = tokenized_text['input_ids'][:, 1:]
        labels[:, -1] = tokenizer.eos_token_id

        # Store in batch
        batch['input_ids'].append(tokenized_text['input_ids'].squeeze(0))
        batch['labels'].append(labels.squeeze(0))
        batch['loss_mask'].append(loss_mask.squeeze(0))

    # Pad sequences to longest in batch
    batch['input_ids'] = pad_sequence(
        batch['input_ids'],
        batch_first=True,
        padding_value=tokenizer.pad_token_id
    )
    batch['labels'] = pad_sequence(
        batch['labels'],
        batch_first=True,
        padding_value=tokenizer.pad_token_id  
    )
    batch['loss_mask'] = pad_sequence(
        batch['loss_mask'],
        batch_first=True,
        padding_value=0
    )
    # batch['input_ids'] = torch.stack(batch['input_ids'])
    # batch['labels'] = torch.stack(batch['labels'])
    # batch['loss_mask'] = torch.stack(batch['loss_mask'])
    # # Convert to BatchEncoding
    return BatchEncoding(batch)


def convert_dataset_to_dataloader(dataset, tokenizer, labels):
    # Convert HF Dataset to PyTorch DataLoader

    dataloader = torch.utils.data.DataLoader(dataset, batch_size=config["Dataset"]["batch_size"], num_workers=config["Dataset"]["num_workers"], shuffle=config["Dataset"]["shuffle"], drop_last=config["Dataset"]["drop_last"], pin_memory=config["Dataset"]["pin_memory"], persistent_workers=config["Dataset"]["persistent_workers"], collate_fn = lambda batch: sft_collate_fn(batch, tokenizer, config, labels))
  
    return dataloader

class SFTDataset(Dataset):
    def __init__(self, dataset_path):
        self.dataset_path = dataset_path
        self.dataset = _load_dataset(dataset_path)
    

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx]















from typing import Dict
import wandb
from smolhub.helper.dataset.load_config import Config

config = Config().get_config()

class Visualizer:
    def __init__(self):
        wandb.init(project=config["wandb"]["project_name"])

    def log(self, metrics: Dict[str, float]):
        wandb.log(metrics)

    def close(self):
        wandb.finish()

    